from __future__ import print_function

import itertools
from math import ceil

from sympy.utilities.iterables import partitions, permutations, multiset_partitions

# pylint does not know sage
from sage.graphs.graph import Graph  # pylint: disable=import-error
# sage documentation says compositions are better than OrderedPartitions,
# no idea why....
from sage.combinat.composition import Compositions as sage_part  # pylint: disable=import-error
from sage.combinat.partition import OrderedPartitions  # pylint: disable=import-error
from sage.misc.cachefunc import cached_function  # pylint: disable=import-error

import admcycles.diffstrata.levelgraph
from admcycles.diffstrata.sig import Signature

############################################################
############################################################
# Old BIC generation:
# This was the first attempt at generating BICs
# (generating all stable graphs with admcycles and then
# putting all possible bipartite structures etc on them).
# It's nice for historic reasons but painfully slow!
# Definitely use bic_alt below instead!!!
############################################################
############################################################


def SageGraph(gr):
    # converts stgraph gr in Sage graph
    legdic = {i: vnum for vnum in range(
        len(gr.legs(copy=False))) for i in gr.legs(vnum, copy=False)}
    return Graph([(legdic[a], legdic[b])
                  for (a, b) in gr.edges(copy=False)], loops=True, multiedges=True)


def bics(g, orders):
    """
    Generate BICs for stratum with signature orders of genus g.

    DO NOT USE!!! USE bic_alt instead!!!

    Args:
        g (int): genus
        orders (tuple): signature tuple

    Returns:
        list: list of BICs
    """
    print('WARNING! This is not the normal BIC generation algorithm!')
    print("Only use if you know what you're doing!")
    n = len(orders)
    bound = g + n
    result = []  # list of levelgraphs
    orderdict = {i + 1: orders[i] for i in range(n)}

    for e in range(1, bound + 1):
        # look at which stgraphs in Mbar_g,n with e edges are bipartite
        for gr in new_list_strata(g, n, e):
            check = SageGraph(gr).is_bipartite(True)
            if check[0]:
                # check[1] contains dictionary {vertices: 0 or 1}
                vert1 = [v for v in check[1] if check[1][v] == 0]
                vert2 = [v for v in check[1] if check[1][v] == 1]
                result += bics_on_bipgr(gr, vert1, vert2, orderdict)
                result += bics_on_bipgr(gr, vert2, vert1, orderdict)
    return result


def bics_on_bipgr(gr, vertup, vertdown, orderdict):
    # takes stable graph and partition of range(len(gr.genera)) into vertup, vertdown
    # creates list of possible pole orders at edges
    result = []

    # removed from admcycles so reinserted here for legacy purposes:
    def dicunion(*dicts):
        return dict(itertools.chain.from_iterable(dct.items()
                                                  for dct in dicts))

    numvert = len(gr.genera(copy=False))
    halfedges = gr.halfedges()
    halfedgeinvers = {e0: e1 for (e0, e1) in gr.edges(copy=False)}
    halfedgeinvers.update({e1: e0 for (e0, e1) in gr.edges(copy=False)})

    levels = [0 for i in range(numvert)]
    for i in vertdown:
        levels[i] = -1

    helist = [[l for l in ll if l in halfedges]
              for ll in gr.legs(copy=False)]  # list of half-edges of each vertex
    # list (with len numvert) of orders that need to be provided by pole
    # orders of half-edges (not legs)
    degs = []
    for v in range(numvert):
        degs.append(2 * gr.genera(v) - 2 - sum([orderdict[l]
                                                for l in gr.list_markings(v)]))
        # if degs<0:
        #     return []

    weightparts = []
    for v in vertup:
        vweightpart = []
        for part in OrderedPartitions(degs[v] + len(helist[v]), len(helist[v])):
            # dictionary of weights from edges connected to vertex v
            vdic = {helist[v][i]: part[i] - 1 for i in range(len(helist[v]))}
            vdic.update(
                {halfedgeinvers[helist[v][i]]: -part[i] - 1 for i in range(len(helist[v]))})
            # print vdic
            vweightpart.append(vdic)
        weightparts += [vweightpart]
        # print weightparts

    for parts in itertools.product(*weightparts):
        # print parts
        poleorders = dicunion(orderdict, *parts)
        CandGraph = admcycles.diffstrata.levelgraph.LevelGraph(
            gr.genera(
                copy=False), gr.legs(
                copy=False), gr.edges(
                copy=False), poleorders, levels, quiet=True)
        if CandGraph.is_legal(True) and CandGraph.checkadmissible(True):
            result.append(CandGraph)
    return result


def new_list_strata(g, n, r):
    return [lis[0] for lis in admcycles.admcycles.degeneration_graph(int(g), n, r)[
        0][r]]

############################################################
############################################################

############################################################
############################################################
# New BIC generation. This is currently used.
############################################################
############################################################


@cached_function
def bic_alt_noiso(sig):
    """
    Construct all non-horizontal divisors in the stratum sig.

    More precisely, each BIC is LevelGraph with two levels numbered 0, -1
    and marked points 1,...,n where the i-th point corresponds to the element
    i-1 of the signature.

    Note that this is the method called by GeneralisedStratum.gen_bic.

    Args:
        sig (Signature): Signature of the stratum.

    Returns:
        list: list of 2-level non-horizontal LevelGraphs.

    EXAMPLES::

        sage: from admcycles.diffstrata import *
        sage: assert comp_list(bic_alt(Signature((1,1))),\
        [LevelGraph([1, 0],[[3, 4], [1, 2, 5, 6]],[(3, 5), (4, 6)],{1: 1, 2: 1, 3: 0, 4: 0, 5: -2, 6: -2},[0, -1],True),\
        LevelGraph([1, 1, 0],[[4], [3], [1, 2, 5, 6]],[(3, 5), (4, 6)],{1: 1, 2: 1, 3: 0, 4: 0, 5: -2, 6: -2},[0, 0, -1],True),\
        LevelGraph([1, 1],[[3], [1, 2, 4]],[(3, 4)],{1: 1, 2: 1, 3: 0, 4: -2},[0, -1],True),\
        LevelGraph([2, 0],[[3], [1, 2, 4]],[(3, 4)],{1: 1, 2: 1, 3: 2, 4: -4},[0, -1],True)])

        sage: assert comp_list(bic_alt(Signature((2,))),\
        [LevelGraph([1, 1],[[2], [1, 3]],[(2, 3)],{1: 2, 2: 0, 3: -2},[0, -1],True),\
        LevelGraph([1, 0],[[2, 3], [1, 4, 5]],[(2, 4), (3, 5)],{1: 2, 2: 0, 3: 0, 4: -2, 5: -2},[0, -1],True)])

        sage: assert comp_list(bic_alt(Signature((4,))),\
        [LevelGraph([1, 1, 0],[[2, 4], [3], [1, 5, 6, 7]],[(2, 5), (3, 6), (4, 7)],{1: 4, 2: 0, 3: 0, 4: 0, 5: -2, 6: -2, 7: -2},[0, 0, -1],True),\
        LevelGraph([1, 1, 1],[[3], [2], [1, 4, 5]],[(2, 4), (3, 5)],{1: 4, 2: 0, 3: 0, 4: -2, 5: -2},[0, 0, -1],True),\
        LevelGraph([2, 0],[[2, 3], [1, 4, 5]],[(2, 4), (3, 5)],{1: 4, 2: 2, 3: 0, 4: -4, 5: -2},[0, -1],True),\
        LevelGraph([2, 0],[[2, 3], [1, 4, 5]],[(2, 4), (3, 5)],{1: 4, 2: 1, 3: 1, 4: -3, 5: -3},[0, -1],True),\
        LevelGraph([1, 0],[[2, 3, 4], [1, 5, 6, 7]],[(2, 5), (3, 6), (4, 7)],{1: 4, 2: 0, 3: 0, 4: 0, 5: -2, 6: -2, 7: -2},[0, -1],True),\
        LevelGraph([1, 2],[[2], [1, 3]],[(2, 3)],{1: 4, 2: 0, 3: -2},[0, -1],True),\
        LevelGraph([2, 1],[[2], [1, 3]],[(2, 3)],{1: 4, 2: 2, 3: -4},[0, -1],True),\
        LevelGraph([1, 1],[[2, 3], [1, 4, 5]],[(2, 4), (3, 5)],{1: 4, 2: 0, 3: 0, 4: -2, 5: -2},[0, -1],True)])

        sage: len(bic_alt(Signature((1,1,1,1)))) # long time (2 seconds)
        102

        sage: len(bic_alt(Signature((2,2,0,-2))))
        61

        sage: len(bic_alt((2,2,0,-2)))
        61
    """
    if isinstance(sig, tuple):
        sig = Signature(sig)
    # for keeping track of the numbered half-edges, we remember the index
    zeros = [i + 1 for i, a in enumerate(sig.sig) if a > 0]
    z = sig.z
    poles = [i + 1 for i, a in enumerate(sig.sig) if a < 0]
    p = sig.p
    marked_points = [i + 1 for i, a in enumerate(sig.sig) if a == 0]
    mp = len(marked_points)
    g = sig.g
    orders = {i + 1: ord for i, ord in enumerate(sig.sig)}

    n = sig.n

    if sig.k == 1:
        # As every g=0 top component needs at least one pole, the bottom max
        # depends on this:
        if p > 0:
            g_bot_max = g
            g_top_min = 0
            pole_ind = [0, 1]
        else:
            g_bot_max = g - 1
            g_top_min = 1
            pole_ind = [0]  # poles don't need to be distributed...
    elif sig.k == 2:
        # For quadratic differentials, g=0 top components without poles are
        # possible.
        g_bot_max = g
        g_top_min = 0
        pole_ind = [0, 1] if p > 0 else [0]
    else:
        raise ValueError("Not implemented for k>2")

    only_even = sig.k == 2 and sig.difftype == "gs"

    found = []

    for bot_comp_len in range(1, z + mp + 1):
        # every component on bottom level needs at least one zero or two marked points
        # (in case it's genus 0 and has only one edge with a pole of order -2k going up...)
        # Therefore, we split the zeros into top and bottom and distribute
        # these:
        for parts in itertools.chain(
                multiset_partitions(zeros, 2), iter([[zeros, []]])):
            for i in [0, 1]:
                if mp > 0 or len(parts[i]) >= bot_comp_len:
                    bottom_zeros = parts[i]
                    top_zeros = parts[1 - i]
                else:
                    continue
                # if there are no marked points, every component needs a zero
                # otherwise we are more flexible (this could be improved, but
                # is good enoug for now)
                if mp == 0:
                    bot_zero_gen = _distribute_fully(
                        bottom_zeros, bot_comp_len)
                else:
                    bot_zero_gen = _distribute_points(
                        bottom_zeros, bot_comp_len)
                for distr_bot_zeros in bot_zero_gen:
                    # partition genus into top, bottom and graph:
                    for total_g_bot in range(g_bot_max + 1):
                        for total_g_top in range(
                                g_top_min, g - total_g_bot + 1):
                            total_g_graph = g - total_g_bot - total_g_top
                            # partition bottom genus into components
                            for distr_bot_g in _distribute_part_ordered(
                                    total_g_bot, bot_comp_len):
                                # first test:
                                # orders on each component must add up to k(2g-2), from now on only poles get
                                # added and we need at least one edge going up (adding at least a pole of order -(k+1))
                                # So if sum(zeros) < k(2g-2)+(k+1), we're
                                # screwed
                                if not all(sum(orders[bz] for bz in distr_bot_zeros[c]) >= sig.k * (
                                        distr_bot_g[c] - 2) + sig.k + 1 for c in range(bot_comp_len)):
                                    continue
                                # start distributing poles, first between top
                                # and bottom:
                                for pole_parts in itertools.chain(
                                        multiset_partitions(poles, 2), iter([[poles, []]])):
                                    for ip in pole_ind:  # no poles => nothing to distribute
                                        bottom_poles = pole_parts[ip]
                                        top_poles = pole_parts[1 - ip]
                                        # if k == 1, poles on top necessary for g=0 components
                                        if sig.k == 1 and total_g_top == 0 and len(
                                                top_poles) == 0:
                                            continue
                                        # poles are optional (i.e. not every vertex needs a pole)
                                        for distr_bot_poles in _distribute_points(
                                                bottom_poles, bot_comp_len):
                                            # we save the "space" left to distribute among poles from
                                            # half-edges:
                                            spaces_bot = [(-sum(orders[tz] for tz in distr_bot_zeros[c])
                                                           - sum(orders[tp] for tp in distr_bot_poles[c])
                                                           + sig.k * (2 * distr_bot_g[c] - 2)) for c in range(bot_comp_len)]
                                            # retest: each space must be at least -(k+1)
                                            if not all(s <= -(sig.k + 1)
                                                       for s in spaces_bot):
                                                continue
                                            # get the total orders of poles on the top components.
                                            # Note that for k>1 there are poles coming from the edges.
                                            if sig.k == 1:
                                                # the total order of all poles on the top components
                                                max_total_pole_order_top = sum(
                                                    orders[p] for p in top_poles)
                                                # the maximal number of poles on the top components
                                                max_poles_top = len(top_poles)
                                            elif sig.k == 2:
                                                # we get poles of order -1 from edges at the top iff we can
                                                # add edges with poles of order -3 at the bottom end
                                                max_poles_from_edges = sum(
                                                    [(-s / 3) if s % 3 == 0 else (int(-s / 3) - 1) for s in spaces_bot])
                                                max_total_pole_order_top = sum(
                                                    orders[p] for p in top_poles) - max_poles_from_edges
                                                max_poles_top = len(
                                                    top_poles) + max_poles_from_edges
                                            else:
                                                raise ValueError(
                                                    "Not implemented for k>2")
                                            # each g=0 component on top level needs at least poles of total order -2*k
                                            max_g0_top = min(
                                                int(-max_total_pole_order_top / (2 * sig.k)), max_poles_top)
                                            # iterate over top components
                                            for top_comp_len in range(
                                                    1, total_g_top + max_g0_top + 1):
                                                # now we know all vertices and the genus distribution, we know the number of edges:
                                                num_of_edges = top_comp_len + bot_comp_len + total_g_graph - 1
                                                # "global" condition on bottom: for each edge there will be
                                                # at least another pole of order k+1
                                                if (sum(sum(orders[bz] for bz in distr_bot_zeros[c])
                                                        + sum(orders[bp] for bp in distr_bot_poles[c])
                                                        for c in range(bot_comp_len))
                                                        - (sig.k + 1) * num_of_edges < sig.k * (2 * total_g_bot - 2 * bot_comp_len)):
                                                    continue
                                                # distribute genus and poles
                                                for distr_top_g in _distribute_part_ordered(
                                                        total_g_top, top_comp_len):
                                                    for distr_top_poles in _distribute_points(
                                                            top_poles, top_comp_len):
                                                        # test: if orders add up to more than k(2g-2) this won't work
                                                        # (effectively this should only concern g=0 components here...)
                                                        if not all(sum(orders[tp] for tp in distr_top_poles[c])
                                                                   # Edges may give additional poles if k > 1
                                                                   + (1 - sig.k) * \
                                                                   num_of_edges
                                                                   <= sig.k * (2 * distr_top_g[c] - 2)
                                                                   for c in range(top_comp_len)):
                                                            continue
                                                        # distribute remaining zeros optionally
                                                        for distr_top_zeros in _distribute_points(
                                                                top_zeros, top_comp_len):
                                                            # again, we record the spaces:
                                                            spaces_top = [-sum(orders[tp] for tp in distr_top_poles[c])
                                                                          - sum(orders[tz] for tz in distr_top_zeros[c])
                                                                          + sig.k *
                                                                          (2 *
                                                                           distr_top_g[c] - 2)
                                                                          for c in range(top_comp_len)]  # yapf: disable
                                                            # retest:
                                                            if not all(
                                                                    s >= (1 - sig.k) * num_of_edges for s in spaces_top):
                                                                continue
                                                            # We distribute the edges together with their prongs/poleorders
                                                            # before we check for connectednes
                                                            # We start on bottom level, because there each half-edge
                                                            # comes with a mandatory pole:
                                                            for half_edges_bot, orders_he in _place_legs_on_bot(
                                                                    spaces_bot, num_of_edges, n + num_of_edges, sig.k, only_even):
                                                                # note for the numbering of half-edges:
                                                                # * the HE on bot are numbered n+num_of_edges+1,...,n+2*num_of_edges
                                                                # * the HE on top are numbered the same for now, i.e. edges are (l,l) ,
                                                                #   (and will be renumbered n+1,...,n+num_of_edges in a moment)
                                                                for half_edges_top in _place_legs_on_top(
                                                                        spaces_top, orders_he, sig.k):
                                                                    # check if graph is connected
                                                                    if not is_connected(
                                                                            half_edges_bot, half_edges_top):
                                                                        continue
                                                                    # add in the orders for top edges (note the offset)
                                                                    for c in half_edges_bot:
                                                                        for l in c:
                                                                            orders_he[l - num_of_edges] = - \
                                                                                2 * sig.k - orders_he[l]
                                                                    genera = distr_top_g + distr_bot_g
                                                                    # Distribute order 0 marked points
                                                                    for distr_mp in _distribute_points(
                                                                            marked_points, len(genera)):
                                                                        # combine nested lists of legs:
                                                                        legs = ([distr_top_zeros[c] + distr_top_poles[c]
                                                                                 + distr_mp[c]
                                                                                 # renumber top HE
                                                                                 + [l - num_of_edges for l in half_edges_top[c]]
                                                                                 for c in range(top_comp_len)]  # top component
                                                                                + [distr_bot_zeros[c] + distr_bot_poles[c]
                                                                                    # offset
                                                                                    + distr_mp[c + top_comp_len]
                                                                                    + half_edges_bot[c]
                                                                                    for c in range(bot_comp_len)]  # bot component
                                                                                )  # yapf: disable
                                                                        # check stability:
                                                                        if any(
                                                                                len(ls) < 3 for c, ls in enumerate(legs) if genera[c] == 0):
                                                                            continue
                                                                        lg_data = (genera, legs,
                                                                                   [(l, l + num_of_edges) for l in range(
                                                                                       n + 1, n + num_of_edges + 1)],
                                                                                   # orders as dict
                                                                                   _merge_dicts(
                                                                                       orders, orders_he),
                                                                                   # levels
                                                                                   [0] * top_comp_len + \
                                                                                   [-1] * \
                                                                                   bot_comp_len,
                                                                                   # True
                                                                                   # #
                                                                                   # quiet
                                                                                   )
                                                                        if sig.k == 1:
                                                                            LG = admcycles.diffstrata.levelgraph.LevelGraph(
                                                                                *lg_data)
                                                                        elif sig.k == 2:
                                                                            LG = admcycles.diffstrata.quadraticlevelgraph.QuadraticLevelGraph(
                                                                                *lg_data)
                                                                        else:
                                                                            raise ValueError(
                                                                                "Not implemented for k>2")
                                                                        if LG.checkadmissible(
                                                                                quiet=True):
                                                                            if sig.k == 1:
                                                                                if LG.is_legal(
                                                                                        quiet=True):
                                                                                    found.append(
                                                                                        LG)
                                                                            elif sig.k == 2:
                                                                                if (sig.difftype == "gs" and LG.is_legal_global_square(quiet=True)) \
                                                                                        or (sig.difftype == "p" and LG.is_legal_primitive(quiet=True)):
                                                                                    found.append(
                                                                                        LG)
                                                                            else:
                                                                                raise ValueError(
                                                                                    "Not implemented for k>2")
                                                                        else:
                                                                            # this should not happen!
                                                                            print(
                                                                                "Not admissible(!): ", LG)
    found = list(set(found))  # remove duplicates
    return found


def bic_alt(sig):
    """
    The BICs of the stratum sig up to isomorphism of LevelGraphs.

    This should not be used directly, use a GeneralisedStratum or Stratum
    instead to obtain EmbeddedLevelGraphs and the correct isomorphism classes.

    Args:
        sig (tuple): signature tuple

    Returns:
        list: list of LevelGraphs.
    """
    return isom_rep(bic_alt_noiso(sig))  # remove duplicates up to iso


def _merge_dicts(x, y):
    z = x.copy()
    z.update(y)
    return z


def _place_legs_on_bot(space, num_of_points, start, k, only_even):
    # iterator distributing n legs with pole orders on (bottom) components according to space (each pole order is at least -(k+1))
    # here space is a list of k(2g-2)-stuff we already placed on each bottom component
    # return a pair (pointlist,orderdict) where
    # * pointlist is a nested list of points (numbered start+1,...,n)
    # * orderdict is a dictionary leg number : order
    legal_splits = []
    distr = []  # list to hold the current distribution

    def split(n):
        # distribute n points onto space and add to distr
        # note that space and distr are accessed from back to front...
        if n == 0:
            # done
            # correct signs and reverse order
            legal_splits.append([[-a for a in c] for c in reversed(distr)])
            return
        else:
            if not space:  # check if empty
                return
            current = space.pop()
            remaining_comp = len(space)
            # there is a sign issue here: partitions are of positive integers, our pole orders are negative
            # each leg has pole order at least k+1, so we only list these partitions
            # moreover, we may place at most n-remaining_comp points on this
            # component
            if only_even:
                possibilities = [[c * 2 for c in pos]
                                 for pos in sage_part(-current / 2, min_part=ceil((k + 1) / 2.), max_length=n - remaining_comp)]
            else:
                possibilities = sage_part(-current, min_part=k + 1,
                                          max_length=n - remaining_comp).list()
            if possibilities:  # check if non-empty
                for possibility in possibilities:
                    distr.append(possibility)
                    # recursion
                    split(n - len(possibility))
                    # undo damage
                    distr.pop()
            space.append(current)
    # generate splits:
    split(num_of_points)
    # we now convert the generated lists (of pole orders) into the format: nested list of numbered legs
    # Note:
    # * the legs are numbered starting from start+1 to start+num_of_points+1
    # * the orders are saved in the list of dictionaries edge_orders_bot
    for dist in legal_splits:
        order_dic = {}
        p = start + 1
        for c in range(len(dist)):
            for a in range(len(dist[c])):
                order_dic[p] = dist[c][a]
                dist[c][a] = p
                p += 1
        yield (dist, order_dic)


def _place_legs_on_top(space, orders_bot, k):
    # iterator distributing n legs with zero orders (determined by their friends on the bottom component)
    # onto the top components according to space
    # here space is a list of k(2g_comp-2)-stuff we already placed on each top component
    # return a pointlist (numbered according to keys of orders_bot for now...)
    legal_splits = []
    distr = [[] for _ in space]  # list to hold current distribution
    # we sort the points by order, as the ones with the highest order are the hardest to place
    # note that this is actually reverse sorted, as the top order is -2k-bottom order, but this
    # is good, as we want to manipulate the list from the end
    ordered_keys = [l for l, v in reversed(
        sorted(orders_bot.items(), key=lambda o: o[1]))]

    def splits(keys):
        # distribute the points (keys) onto spaces
        if not keys:
            # done if we hit all components and all spaces are 0
            if all(distr) and all(s == 0 for s in space):
                legal_splits.append([[a for a in c] for c in distr])
            return
        else:
            # check if there are enough points left
            # components that don't have a point yet
            remaining_comp = len([hit for hit in distr if not hit])
            remaining_pole_order = 0 if k == 1 else sum(
                o for o in [-2 * k - orders_bot[l] for l in keys] if o < 0)
            if remaining_comp > len(keys):
                return
            current = keys.pop()
            current_order = -2 * k - orders_bot[current]
            # try to place current on all components
            for i in range(len(space)):
                if space[i] >= current_order + remaining_pole_order:
                    space[i] -= current_order
                    distr[i].append(current)
                    splits(keys)  # recursion
                    # undo changes:
                    space[i] += current_order
                    distr[i].pop()
            keys.append(current)
    # generate splits:
    splits(ordered_keys)
    return legal_splits


def _distribute_fully(points, n):
    # iterator distributing list of points onto n components, where each component gets at least one point
    # return a nested list of points
    # generate partitions into n subsets of permuted_points:
    for part in multiset_partitions(points, n):
        # we need to consider all permutations (multiset_partitions sorts!)
        for permuted_points in permutations(part):
            yield list(permuted_points)  # permutations give tuples...


def _b_ary_gen(x, b, n):
    # generator for reverse b-ary integers x of length n
    for _ in itertools.repeat(None, n):
        r = x % b
        x = (x - r) // b
        yield r


def _distribute_points(points, n):
    # iterator distributing list of points onto n components (some might not receive any)
    # return a nested list of points
    l = len(points)
    # n^l possibilities:
    for i in range(n**l):
        point_list = [[] for j in range(n)]
        # n-ary representation of i tells us where the points should go
        for pos, d in enumerate(_b_ary_gen(i, n, l)):
            point_list[d].append(points[pos])
        yield point_list


def _distribute_part_ordered(g, n):
    # iterator of partitions (g_1,...,g_k) of g of length k <= n distributed onto n points
    # return a list of length n (that sums to g)
    if n < g:
        maxi = n
    else:
        maxi = None
    for part_dict in partitions(g, m=maxi):
        # partitions are dictionaries {g_i : multiplicity}
        part = []
        for k in part_dict.keys():
            part += [k] * part_dict[k]
        # fill up with zeros:
        part += (n - len(part)) * [0]
        yield part
        # we do not have to permute if everything else is permuted....
        # for perm_part in set(permutations(part)):
        #     yield perm_part


def isom_rep(L):
    """
    Return a list of representatives of isomorphism classes of L.

    TODO: optimise!
    """
    dist_list = []
    for g in L:
        if all(not g.is_isomorphic(h) for h in dist_list):
            dist_list.append(g)
    return dist_list


def comp_list(L, H):
    r"""
    Compare two lists of LevelGraphs (up to isomorphism).

    Returns a tuple: (list L without H, list H without L)
    """
    return ([g for g in L if not any(g.is_isomorphic(h) for h in H)],
            [g for g in H if not any(g.is_isomorphic(h) for h in L)])


def is_connected(half_edges_top, half_edges_bottom):
    """
    Check if graph given by the two sets of half edges is connected.
    We do this by depth-first search.
    Note that we assume that both ends of each edge have the same number,
    i.e. each edge is of the form (j,j).

    EXAMPLES::

        sage: from admcycles.diffstrata.bic import is_connected
        sage: is_connected([[1],[2]],[[1,2]])
        True
        sage: is_connected([[1],[2]],[[2],[1]])
        False
    """
    def _connected_comp(current_edges, other_edges,
                        seen_current, seen_other, current_vert=0):
        seen_current.append(current_vert)
        for e in current_edges[current_vert]:
            for (i, edges) in enumerate(other_edges):
                if e in edges:
                    if i not in seen_other:
                        _connected_comp(other_edges, current_edges,
                                        seen_other, seen_current, i)
                    break
        return (seen_current, seen_other)
    seen_top = []
    seen_bottom = []
    _connected_comp(half_edges_top, half_edges_bottom, seen_top, seen_bottom)
    return len(seen_top) == len(half_edges_top) and len(
        seen_bottom) == len(half_edges_bottom)


def test_bic_algs(sig_list=None):
    """
    Compare output of bics and bic_alt.

    EXAMPLES::

        sage: from admcycles.diffstrata import *
        sage: test_bic_algs()  # long time (45 seconds)  # skip, not really needed + long # doctest: +SKIP
        (1, 1): ([], [])
        (1, 1, 0, 0, -2): ([], [])
        (2, 0, -2): ([], [])
        (1, 0, 0, 1): ([], [])
        (1, -2, 2, 1): ([], [])
        (2, 2): ([], [])

        sage: test_bic_algs([(0,0),(2,1,1)])  # long time (50 seconds)  # doctest: +SKIP
        (0, 0): ([], [])
        (2, 1, 1): ([], [])

        sage: test_bic_algs([(1,0,-1),(2,),(4,),(1,-1)])
        WARNING! This is not the normal BIC generation algorithm!
        Only use if you know what you're doing!
        (1, 0, -1): ([], [])
        WARNING! This is not the normal BIC generation algorithm!
        Only use if you know what you're doing!
        (2,): ([], [])
        WARNING! This is not the normal BIC generation algorithm!
        Only use if you know what you're doing!
        (4,): ([], [])
        WARNING! This is not the normal BIC generation algorithm!
        Only use if you know what you're doing!
        (1, -1): ([], [])
    """
    if sig_list is None:
        sig_list = [(1, 1), (1, 1, 0, 0, -2), (2, 0, -2),
                    (1, 0, 0, 1), (1, -2, 2, 1), (2, 2)]
    for sig in sig_list:
        Sig = Signature(sig)
        print(
            "%r: %r" %
            (Sig.sig,
             comp_list(
                 bics(
                     Sig.g,
                     Sig.sig),
                 bic_alt(Sig))))



from __future__ import absolute_import
from __future__ import print_function

# pylint does not know sage
from sage.structure.sage_object import SageObject  # pylint: disable=import-error
from sage.misc.cachefunc import cached_method  # pylint: disable=import-error
from sage.rings.rational_field import QQ  # pylint: disable=import-error

import admcycles.admcycles
import admcycles.stratarecursion

import admcycles.diffstrata.elgtautclass

from admcycles.diffstrata.auxiliary import hash_AG


class AdditiveGenerator (SageObject):
    """
    Product of Psi classes on an EmbeddedLevelGraph (of a stratum X).

    The information of a product of psi-class on an EmbeddedLevelGraph, i.e. a
    leg_dict and an enhanced_profile, where leg_dict is a dictionary on the legs
    leg -> exponent of the LevelGraph associated to the enhanced profile, i.e.
    (profile,index) or None if we refer to the class of the graph.

    We (implicitly) work inside some stratum X, where the enhanced profile
    makes sense.

    This class should be considered constant (hashable)!
    """

    def __init__(self, X, enh_profile, leg_dict=None):
        """
        AdditiveGenerator for psi polynomial given by leg_dict on graph
        corresponding to enh_profile in X.

        Args:
            X (GeneralisedStrataum): enveloping stratum
            enh_profile (tuple): enhanced profile (in X)
            leg_dict (dict, optional): dictionary leg of enh_profile -> exponent
                encoding a psi monomial. Defaults to None.
        """
        self._X = X
        self._hash = hash_AG(leg_dict, enh_profile)
        self._enh_profile = (tuple(enh_profile[0]), enh_profile[1])
        self._leg_dict = leg_dict
        self._G = self._X.lookup_graph(*enh_profile)
        # dictionary leg -> level
        # Careful! These are leg numbers on the whole graph, not on
        # the graphs inside the LevelStrata!!
        self._level_dict = {}
        if leg_dict is not None:
            for l in leg_dict:
                self._level_dict[l] = self._G.LG.level_number(
                    self._G.LG.levelofleg(l))
        self._inv_level_dict = {}
        for leg in self._level_dict:
            try:
                self._inv_level_dict[self._level_dict[leg]].append(leg)
            except KeyError:
                self._inv_level_dict[self._level_dict[leg]] = [leg]

    @classmethod
    def from_hash(cls, X, hash):
        """
        AdditiveGenerator from a hash generated with hash_AG.

        Args:
            X (GeneralisedStratum): Enveloping stratum.
            hash (tuple): hash from hash_AG

        Returns:
            AdditiveGenerator: AG from hash.
        """
        if hash[0] is None:
            leg_dict = None
        else:
            leg_dict = dict(hash[0])
        return cls(X, (hash[1], hash[2]), leg_dict)

    def __hash__(self):
        return hash(self._hash)

    def __eq__(self, other):
        try:
            return self._hash == other._hash
        except AttributeError:
            return NotImplemented

    def __repr__(self):
        return "AdditiveGenerator(X=%r,enh_profile=%r,leg_dict=%r)"\
            % (self._X, self._enh_profile, self._leg_dict)
        # Better, but destroys tests:
        # return "AdditiveGenerator(enh_profile=%r,leg_dict=%r)"\
        #      % (self._enh_profile, self._leg_dict)

    def __str__(self):
        str = ""
        if self._leg_dict is not None:
            for l in self._leg_dict:
                str += "Psi class %r with exponent %r on level %r * "\
                    % (l, self._leg_dict[l], self._level_dict[l])
        str += "Graph %r" % (self._enh_profile,)
        return str

    def __mul__(self, other):
        """
        Multiply to psi products on the same graph (add dictionaries).

        Args:
            other (AdditiveGenerator): Product of psi classes on same graph.

        Returns:
            AdditiveGenerator: Product of psi classes on same graph.

        EXAMPLES::


            Also works without legs.

        """
        # Check that other is an AdditiveGenerator for the same graph:
        try:
            if self._X != other._X or self._enh_profile != other._enh_profile:
                return NotImplemented
            other_leg_dict = other._leg_dict
        except AttributeError:
            return NotImplemented
        # "unite" the leg_dicts:
        if self._leg_dict is None:
            self_leg_dict = {}
        else:
            self_leg_dict = self._leg_dict
        if other_leg_dict is None:
            other_leg_dict = {}
        new_leg_dict = {l: self_leg_dict.get(l, 0) + other_leg_dict.get(l, 0)
                        for l in set(self_leg_dict) | set(other_leg_dict)}
        return self._X.additive_generator(self._enh_profile, new_leg_dict)

    def __rmul__(self, other):
        self.__mul__(other)

    def __pow__(self, n):
        return self.pow(n)

    @property
    def enh_profile(self):
        return self._enh_profile

    @property
    def psi_degree(self):
        """
        Sum of powers of psi classes of self.
        """
        if self._leg_dict is None:
            return 0
        else:
            return sum(self._leg_dict.values())

    @cached_method
    def dim_check(self):
        """
        Check if, on any level, the psi degree is higher than the dimension.

        Returns:
            bool: False if the class is 0 for dim reasons, True otherwise.
        """
        # remove if degree > dim(X)
        if self.degree > self._X.dim():
            return False
        if self.codim == 0:
            # Avoid crazy infinite recursion for smooth graph :-)
            return True
        # for each level, check if psi product on level exceeds level dimension
        for level_number in range(self.codim + 1):
            assert self.level_dim(level_number) >= 0
            if self.degree_on_level(
                    level_number) > self.level_dim(level_number):
                return False
        return True

    @property
    def codim(self):
        """
        The codimension of the graph (number of levels)

        Returns:
            int: length of the profile
        """
        return len(self._enh_profile[0])

    @property
    def degree(self):
        """
        Degree of class, i.e. codimension of graph + psi-degree

        Returns:
            int: codim + psi_degree
        """
        # degree = codim of graph + powers of psi classes
        return self.codim + self.psi_degree

    @property
    def leg_dict(self):
        return self._leg_dict

    @property
    def level_dict(self):
        """
        The dictionary mapping leg -> level
        """
        return self._level_dict

    @property
    def inv_level_dict(self):
        """
        The dictionary mapping level -> list of legs on level.

        Returns:
            dict: level -> list of legs.
        """
        return self._inv_level_dict

    @cached_method
    def degree_on_level(self, level):
        """
        Total degree of psi classes on level.

        Args:
            level (int): (relative) level number (i.e. 0...codim)

        Raises:
            RuntimeError: Raised for level number out of range.

        Returns:
            int: sum of exponents of psis appearing on this level.
        """
        if level not in range(self.codim + 1):
            raise RuntimeError(
                "Illegal level number: %r on %r" % (level, self))
        try:
            return sum(self._leg_dict[leg]
                       for leg in self._inv_level_dict[level])
        except KeyError:
            # no psis on this level
            return 0

    def level(self, level_number):
        """
        Level of underlying graph.

        Args:
            level_number (int): (relative) level number (0...codim)

        Returns:
            LevelStratum: Stratum at level level_number of self._G.
        """
        return self._G.level(level_number)

    @cached_method
    def level_dim(self, level_number):
        """
        Dimension of level level_number.

        Args:
            level_number (int): (relative) level number (i.e. 0...codim)

        Returns:
            int: dimension of GeneralisedLevelStratum
        """
        level = self._G.level(level_number)
        return level.dim()

    @property
    def stack_factor(self):
        """
        The stack factor, that is the product of the prongs of the underlying graph
        divided by the product of the ells of the BICs and the automorphisms.

        Returns:
            QQ: stack factor
        """
        try:
            return self._stack_factor
        except AttributeError:
            # to get g_Gamma, we have to take the product of prongs/lcm for
            # each bic:
            prod = 1
            for k in self._G.LG.prongs.values():
                prod *= k

            p, _ = self.enh_profile

            bic_contr = 1
            for i in p:
                bic_contr *= self._X.bics[i].ell

            stack_factor = QQ(prod) / QQ(bic_contr *
                                         len(self._G.automorphisms))

            self._stack_factor = stack_factor
            return self._stack_factor

    @cached_method
    def as_taut(self):
        """
        Helper method, returns [(1,self)] as default input to ELGTautClass.
        """
        return admcycles.diffstrata.elgtautclass.ELGTautClass(self._X, [
                                                              (1, self)])

    @cached_method
    def is_in_ambient(self, ambient_enh_profile):
        """
        Check if ambient_enh_profile is an ambient graph, i.e. self is a degeneration
        of ambient_enh_profile.

        Args:
            ambient_enh_profile (tuple): enhanced profile.

        Returns:
            bool: True if there exists a leg map, False otherwise.

        EXAMPLES::

        """
        return self._X.is_degeneration(self._enh_profile, ambient_enh_profile)

    @cached_method
    def pow(self, n, amb=None):
        """
        Recursively calculate the n-th power of self (in amb), caching all results.

        Args:
            n (int): exponent
            amb (tuple, optional): enhanced profile. Defaults to None.

        Returns:
            ELGTautClass: self^n in CH(amb)
        """
        if amb is None:
            ONE = self._X.ONE
            amb = ((), 0)
        else:
            ONE = self._X.taut_from_graph(*amb)
        if n == 0:
            return ONE
        return self._X.intersection(self.as_taut(), self.pow(n - 1, amb), amb)

    @cached_method
    def exp(self, c, amb=None, stop=None):
        """
        exp(c * self) in CH(amb), calculated via exp_list.

        Args:
            c (QQ): coefficient
            amb (tuple, optional): enhanced profile. Defaults to None.
            stop (int, optional): cut-off. Defaults to None.

        Returns:
            ELGTautClass: the tautological class associated to the
                graded list exp_list.
        """
        # graded pieces are already reduced:
        new_taut_list = []
        for T in self.exp_list(c, amb, stop):
            new_taut_list.extend(T.psi_list)
        return admcycles.diffstrata.elgtautclass.ELGTautClass(
            self._X, new_taut_list, reduce=False)

    @cached_method
    def exp_list(self, c, amb=None, stop=None):
        """
        Calculate exp(c * self) in CH(amb).

        We calculate exp as a sum of powers (using self.pow, i.e. cached)
        and check at each step if the power vanishes (if yes, we obviously stop).

        The result is returned as a list consisting of the graded pieces.

        Optionally, one may specify the cut-off degree using stop (by
        default this is dim + 1).

        Args:
            c (QQ): coefficient
            amb (tuple, optional): enhanced profile. Defaults to None.
            stop (int, optional): cut-off. Defaults to None.

        Returns:
            list: list of ELGTautClasses
        """
        c = QQ(c)
        if amb is None:
            ONE = self._X.ONE
            amb = ((), 0)
        else:
            ONE = self._X.taut_from_graph(*amb)
        e = [ONE]
        f = ONE
        coeff = QQ(1)
        k = QQ(0)
        if stop is None:
            stop = self._X.dim() + 1
        while k < stop and f != self._X.ZERO:
            k += 1
            coeff *= c / k
            f = self.pow(k, amb)
            e.append(coeff * f)
        return e

    def pull_back(self, deg_enh_profile):
        """
        Pull back self to the graph associated to deg_enh_profile.

        Note that this returns an ELGTautClass as there could be several maps.

        More precisely, we return the sum over the pulled back classes divided
        by the number of undegeneration maps.

        Args:
            deg_enh_profile (tuple): enhanced profile of graph to pull back to.

        Raises:
            RuntimeError: raised if deg_enh_profile is not a degeneration of the
                underlying graph of self.

        Returns:
            ELGTautClass: sum of pullbacks of self to deg_enh_profile for each
                undegeneration map divided by the number of such maps.

        """
        if self._leg_dict is None:
            # trivial pullback
            return admcycles.diffstrata.elgtautclass.ELGTautClass(
                self._X, [(1, self._X.additive_generator(deg_enh_profile))])
        else:
            leg_maps = self._X.explicit_leg_maps(
                self._enh_profile, deg_enh_profile)
            if leg_maps is None:
                raise RuntimeError("Pullback failed: %r is not a degeneration of %r")\
                    % (deg_enh_profile, self._enh_profile)
            psi_list = []
            aut_factor = QQ(1) / QQ(len(leg_maps))
            for leg_map in leg_maps:
                new_leg_dict = {leg_map[l]: e for l,
                                e in self._leg_dict.items()}
                psi_list.append(
                    (aut_factor, self._X.additive_generator(
                        deg_enh_profile, new_leg_dict)))
            return admcycles.diffstrata.elgtautclass.ELGTautClass(
                self._X, psi_list)

    def psis_on_level(self, l):
        """
        The psi classes on level l of self.

        Args:
            l (int): level, i.e. 0,...,codim

        Returns:
            dict: psi dictionary on self.level(l).smooth_LG
        """
        L = self.level(l)
        # The psi classes on this level should be expressed in terms of the legs
        # of the smooth_LG of L:
        EG = L.smooth_LG
        try:
            # Careful: the legs of the smooth_LG are numbered 1,...,n
            # The psi classes are still numbered inside the whole graph
            # The conversion runs through the embedding of the LevelStratum
            # and back through the embedding of smooth_LG (dmp_inv)
            psis = {EG.dmp_inv[L.leg_dict[leg]]: self.leg_dict[leg]
                    for leg in self.inv_level_dict[l]}
        except KeyError:
            # no psis on this level
            psis = {}
        return psis

    def evaluate(self, quiet=False, warnings_only=False,
                 admcycles_output=False):
        """
        Evaluate self (cap with the fundamental class of self._X).

        Note that this gives 0 if self is not a top-degree class.

        Evaluation works by taking the product of the evaluation of each level
        (i.e. evaluating, for each level, the psi monomial on this level) and
        multiplying this with the stack factor.

        The psi monomials on the levels are evaluated using admcycles (after
        removing residue conditions).

        Args:
            quiet (bool, optional): No output. Defaults to False.
            warnings_only (bool, optional): Output warnings. Defaults to False.
            admcycles_output (bool, optional): Output admcycles debugging info
                (used when evaluating levels). Defaults to False.

        Raises:
            RuntimeError: raised if there are inconsistencies with the psi
            degrees on the levels.

        Returns:
            QQ: integral of self on X.
        """
        if self.degree < self._X.dim():
            if not quiet or warnings_only:
                print("Warning: %r is not of top degree: %r (instead of %r)" %
                      (self, self.degree, self._X.dim()))
            return 0
        level_list = []
        for l in range(self.codim + 1):
            if self.degree_on_level(l) < self.level_dim(l):
                raise RuntimeError(
                    "%r is of top degree, but not on level %r" % (self, l))
            L = self.level(l)
            value = L.evaluate(
                psis=self.psis_on_level(l),
                quiet=quiet,
                warnings_only=warnings_only,
                admcycles_output=admcycles_output)
            if value == 0:
                return 0
            level_list.append(value)
        # product over levels:
        prod = 1
        for p in level_list:
            prod *= p
        if not quiet:
            print("----------------------------------------------------")
            print("Contribution of Additive generator:")
            print(self)
            print("Product of level-wise integrals: %r" % prod)
            print("Stack factor: %r" % self.stack_factor)
            print("Total: %r" % (prod * self.stack_factor))
            print("----------------------------------------------------")
        return self.stack_factor * prod

    def to_prodtautclass(self):
        """
        Transform self into an admcycles prodtautclass on the underlying stgraph of self.

        Note that this gives the pushforward to M_g,n in the sense that we multiply with
        Strataclass and remove all residue conditions.

        Returns:
            prodtautclass: the prodtautclass of self, multiplied with the Strataclasses of
                the levels and all residue conditions removed.

        EXAMPLES::

            sage: from admcycles.diffstrata import *
            sage: X=Stratum((2,))
            sage: X.additive_generator(((),0)).to_prodtautclass()
            Outer graph : [2] [[1]] []
            Vertex 0 :
            Graph :      [2] [[1]] []
            Polynomial : -7/24*(kappa_1)_0 + 79/24*psi_1
            <BLANKLINE>
            <BLANKLINE>
            Vertex 0 :
            Graph :      [1] [[1, 3, 4]] [(3, 4)]
            Polynomial : -1/48
            <BLANKLINE>
            <BLANKLINE>
            Vertex 0 :
            Graph :      [1, 1] [[3], [1, 4]] [(3, 4)]
            Polynomial : -19/24
            sage: from admcycles.stratarecursion import Strataclass
            sage: X=GeneralisedStratum([Signature((4,-2,-2))], res_cond=[[(0,1)], [(0,2)]])
            sage: (X.additive_generator(((),0)).to_prodtautclass().pushforward() - Strataclass(1, 1, [4,-2,-2], res_cond=[2])).is_zero()
            True

        TESTS::

            sage: from admcycles import diffstrata, psiclass
            sage: X = diffstrata.generalisedstratum.GeneralisedStratum(sig_list = [diffstrata.sig.Signature(tuple([8,-3,-2,-3]))], res_cond = [[(0,1)],[(0,2)]])
            sage: X.psi(1).evaluate()
            9
            sage: v = X.ONE.to_prodtautclass().pushforward()
            sage: (v*psiclass(1,1,4)).evaluate()
            9
        """
        LG = self._G.LG
        stgraph = LG.stgraph
        if any(self.level(l).zeroStratumClass()
               for l in range(self.codim + 1)):
            return admcycles.admcycles.prodtautclass(stgraph, terms=[])  # ZERO
        alpha = []  # prodtautclasses on levels
        vertices = []  # embedding of level into stgraph
        for l in range(self.codim + 1):
            # we pass the psis on this level and hope that in-between terms vanish
            # for dimension reasons:
            psis = self.psis_on_level(l)
            # ELGTautClass on self.level(l) with possibly less RCs
            T = self.level(l).remove_res_cond(psis)
            if T._X.res_cond or len(T.psi_list) > 1:
                alpha.append(T.to_prodtautclass())
            else:
                # Now T is at most a psi product on T._X:
                psis = T.psi_list[0][1].leg_dict
                # self.level(l) has no residue conditions: take prodtautclass
                # of Strataclass
                tautlist = [admcycles.stratarecursion.Strataclass(
                    sig.g, 1, sig.sig) for sig in self.level(l)._sig_list]
                # as it is not ZERO it must be connected!
                assert len(tautlist) == 1
                # we have to include the psi contribution:
                stgraph_level = T._X.smooth_LG.LG.stgraph
                adm_psis = T.psi_list[0][0] * admcycles.admcycles.decstratum(
                    stgraph_level, psi=psis)
                adm_psis_taut = adm_psis.to_tautological_class()  # admcycles.admcycles.tautclass([adm_psis])
                tautlist = [tautlist[0] * adm_psis_taut]
                ptc = admcycles.admcycles.prodtautclass(
                    stgraph_level, protaut=tautlist)
                alpha.append(ptc)
            # Finally, we save the vertices of this level (as vertices of
            # stgraph)
            vertices.append(LG.verticesonlevel(LG.internal_level_number(l)))
        # now we pull back all the alphas to a prodtautclass on stgraph:
        prod = self.stack_factor * admcycles.admcycles.prodtautclass(stgraph)
        for l, ptc in enumerate(alpha):
            prod = prod.factor_pullback(
                vertices[l], ptc)  # returns product (!)
        return prod



r"""
The possible different moduli.
"""

import numbers

# Moduli types
MODULI_SM = 0  # smooth (no edge, just the trivial graph)
MODULI_RT = 1  # rational tails (tree with genus 0 vertices but one)
MODULI_CT = 2  # compact type (tree with any kind of vertices)
MODULI_TL = 3  # tree like (tree with self-loops)
MODULI_ST = 4  # stable curves (general case)

_moduli_to_str = {
    MODULI_SM: 'sm',    # smooth curves
    MODULI_RT: 'rt',    # rational tails
    MODULI_CT: 'ct',    # compact type
    MODULI_TL: 'tl',    # tree like
    MODULI_ST: 'st'     # all stable curves
}
_str_to_moduli = {v: k for k, v in _moduli_to_str.items()}


def get_moduli(arg, default=MODULI_ST, DRpy=False):
    if DRpy:
        return min(3, get_moduli(arg, default))
    if arg is None:
        return default
    elif isinstance(arg, str):
        try:
            return _str_to_moduli[arg]
        except KeyError:
            raise ValueError('invalid moduli {!r}'.format(arg))
    elif isinstance(arg, numbers.Integral):
        if arg < MODULI_SM or arg > MODULI_ST:
            raise ValueError('invalid moduli {!r}'.format(arg))
        return int(arg)
    else:
        raise TypeError("invalid moduli; must be a string 'sm', 'rt', 'ct', 'tl', or 'st' (got {!r})".format(arg))


def socle_degree(g, n, moduli):
    if moduli == MODULI_ST:
        # stable curves
        return 3 * g - 3 + n
    elif moduli == MODULI_TL:
        # tree-like
        # TODO: this is an over estimation. It is very likely that the socle dimension
        # is smaller.
        return 3 * g - 3 + n
    elif moduli == MODULI_CT:
        # compact type
        return 2 * g - 3 + n
    elif moduli == MODULI_RT:
        # rational tails
        return g - 2 + n - (g == 0)
    elif moduli == MODULI_SM:
        # smooth
        return g - 1 + (g == 0) - (n == 0)
    else:
        raise ValueError('unknown moduli')



from __future__ import absolute_import
from __future__ import print_function

from collections import Counter

# pylint does not know sage
from sage.structure.sage_object import SageObject  # pylint: disable=import-error
from sage.misc.cachefunc import cached_method  # pylint: disable=import-error

import admcycles.diffstrata.additivegenerator
import admcycles.admcycles


class ELGTautClass (SageObject):
    """
    A Tautological class of a stratum X, i.e. a formal sum of of psi classes on
    EmbeddedLevelGraphs.

    This is encoded by a list of summands.

    Each summand corresponds to an AdditiveGenerator with coefficient.

    Thus an ELGTautClass is a list with entries tuples (coefficient, AdditiveGenerator).

    These can be added, multiplied and reduced (simplified).

    INPUT :

      * X : GeneralisedStratum that we are on
      * psi_list : list of tuples (coefficient, AdditiveGenerator) as
            described above.
      * reduce=True : call self.reduce() on initialisation
    """

    def __init__(self, X, psi_list, reduce=True):
        self._psi_list = psi_list
        self._X = X
        if reduce:
            self.reduce()

    @classmethod
    def from_hash_list(cls, X, hash_list):
        # This does not reduce!
        return cls(X, [(c, X.additive_generator_from_hash(h))
                       for c, h in hash_list], reduce=False)

    @property
    def psi_list(self):
        return self._psi_list

    def __repr__(self):
        return "ELGTautClass(X=%r,psi_list=%r)"\
            % (self._X, self._psi_list)

    def __str__(self):
        str = "Tautological class on %s\n" % self._X
        for coeff, psi in self._psi_list:
            str += "%s * %s + \n" % (coeff, psi)
        return str

    def __eq__(self, other):
        if isinstance(
                other,
                admcycles.diffstrata.additivegenerator.AdditiveGenerator):
            return self == other.as_taut()
        try:
            return self._psi_list == other._psi_list
        except AttributeError:
            return False

    def __add__(self, other):
        # for sum, we need to know how to add '0':
        if other == 0:
            return self
        try:
            if not self._X == other._X:
                return NotImplemented
            new_psis = self._psi_list + other._psi_list
            return ELGTautClass(self._X, new_psis)
        except AttributeError:
            return NotImplemented

    def __iadd__(self, other):
        return self.__add__(other)

    def __radd__(self, other):
        return self.__add__(other)

    def __neg__(self):
        return (-1) * self

    def __sub__(self, other):
        return self + (-1) * other

    def __mul__(self, other):
        if 0 == other:
            return 0
        elif self._X.ONE == other:
            return self
        # convert AdditiveGenerators to Tautclasses:
        if isinstance(
                other,
                admcycles.diffstrata.additivegenerator.AdditiveGenerator):
            return self * other.as_taut()
        try:
            # check if other is a tautological class
            other._psi_list
        except AttributeError:
            # attempt scalar multiplication:
            new_psis = [(coeff * other, psi) for coeff, psi in self._psi_list]
            return ELGTautClass(self._X, new_psis, reduce=False)
        if not self._X == other._X:
            return NotImplemented
        else:
            return self._X.intersection(self, other)

    def __rmul__(self, other):
        return self.__mul__(other)

    def __pow__(self, exponent):
        if exponent == 0:
            return self._X.ONE
        # TODO: quick check for going over top degree?
        prod = self
        for _ in range(1, exponent):
            prod = self * prod
        return prod

    @cached_method
    def is_equidimensional(self):
        """
        Determine if all summands of self have the same degree.

        Note that the empty empty tautological class (ZERO) gives True.

        Returns:
            bool: True if all AdditiveGenerators in self.psi_list are of same degree,
                False otherwise.
        """
        if self.psi_list:
            first_deg = self.psi_list[0][1].degree
            return all(AG.degree == first_deg for _c, AG in self.psi_list)
        return True

    def reduce(self):
        """
        Reduce self.psi_list by combining summands with the same AdditiveGenerator
        and removing those with coefficient 0 or that die for dimension reasons.
        """
        # we use the hash of the AdditiveGenerators to group:
        hash_dict = Counter()
        for c, AG in self._psi_list:
            hash_dict[AG] += c
        self._psi_list = [(c, AG) for AG, c in hash_dict.items()
                          if c != 0 and AG.dim_check()]

    # To evaluate, we go through the AdditiveGenerators and
    # take the (weighted) sum of the AdditiveGenerators.
    def evaluate(self, quiet=True, warnings_only=False,
                 admcycles_output=False):
        """
        Evaluation of self, i.e. cap with fundamental class of X.

        This is the sum of the evaluation of the AdditiveGenerators in psi_list
        (weighted with their coefficients).

        Each AdditiveGenerator is (essentially) the product of its levels,
        each level is (essentially) evaluated by admcycles.

        Args:
            quiet (bool, optional): No output. Defaults to True.
            warnings_only (bool, optional): Only warnings output. Defaults to False.
            admcycles_output (bool, optional): admcycles debugging output. Defaults to False.

        Returns:
            QQ: integral of self on X

        EXAMPLES::

            sage: from admcycles.diffstrata import *
            sage: X=GeneralisedStratum([Signature((0,0))])
            sage: assert (X.xi^2).evaluate() == 0

            sage: X=GeneralisedStratum([Signature((1,1,1,1,-6))])
            sage: assert set([(X.cnb(((i,),0),((i,),0))).evaluate() for i in range(len(X.bics))]) == {-2, -1}
        """
        if warnings_only:
            quiet = True
        DS_list = []
        for c, AG in self.psi_list:
            value = AG.evaluate(
                quiet=quiet,
                warnings_only=warnings_only,
                admcycles_output=admcycles_output)
            DS_list.append(c * value)
        if not quiet:
            print("----------------------------------------------------")
            print("In summary: We sum")
            for i, summand in enumerate(DS_list):
                print("Contribution %r from AdditiveGenerator" % summand)
                print(self.psi_list[i][1])
                print("(With coefficient %r)" % self.psi_list[i][0])
            print("To obtain a total of %r" % sum(DS_list))
            print("----------------------------------------------------")
        return sum(DS_list)

    def extract(self, i):
        """
        Return the i-th component of self.

        Args:
            i (int): index of self._psi_list

        Returns:
            ELGTautClass: coefficient * AdditiveGenerator at position i of self.
        """
        return ELGTautClass(self._X, [self._psi_list[i]], reduce=False)

    @cached_method
    def degree(self, d):
        """
        The degree d part of self.

        Args:
            d (int): degree

        Returns:
            ELGTautClass: degree d part of self
        """
        new_psis = []
        for c, AG in self.psi_list:
            if AG.degree == d:
                new_psis.append((c, AG))
        return ELGTautClass(self._X, new_psis, reduce=False)

    @cached_method
    def list_by_degree(self):
        """
        A list of length X.dim with the degree d part as item d

        Returns:
            list: list of ELGTautClasses with entry i of degree i.
        """
        deg_psi_list = [[] for _ in range(self._X.dim() + 1)]
        for c, AG in self.psi_list:
            deg_psi_list[AG.degree].append((c, AG))
        return [ELGTautClass(self._X, piece, reduce=False)
                for piece in deg_psi_list]

    def is_pure_psi(self):
        """
        Check if self is ZERO or a psi-product on the stratum.

        Returns:
            boolean: True if self has at most one summand and that is of the form
                AdditiveGenerator(((), 0), psis).

        EXAMPLES::

            sage: from admcycles.diffstrata import *
            sage: X=Stratum((2,))
            sage: X.ZERO.is_pure_psi()
            True
            sage: X.ONE.is_pure_psi()
            True
            sage: X.psi(1).is_pure_psi()
            True
            sage: X.xi.is_pure_psi()
            False
        """
        if not self.psi_list:
            return True
        return len(
            self.psi_list) == 1 and self.psi_list[0][1].enh_profile == ((), 0)

    def to_prodtautclass(self):
        """
        Transforms self into an admcycles prodtautclass on the stable graph of the smooth
        graph of self._X.

        Note that this is essentially the pushforward to M_g,n, i.e. we resolve residues
        and multiply with the correct Strataclasses along the way.

        Returns:
            prodtautclass: admcycles prodtautclass corresponding to self pushed forward
                to the stable graph with one vertex.

        EXAMPLES::

            sage: from admcycles.diffstrata import *
            sage: X=Stratum((2,))
            sage: X.ONE.to_prodtautclass()
            Outer graph : [2] [[1]] []
            Vertex 0 :
            Graph :      [2] [[1]] []
            Polynomial : -7/24*(kappa_1)_0 + 79/24*psi_1
            <BLANKLINE>
            <BLANKLINE>
            Vertex 0 :
            Graph :      [1] [[1, 2, 3]] [(2, 3)]
            Polynomial : -1/48
            <BLANKLINE>
            <BLANKLINE>
            Vertex 0 :
            Graph :      [1, 1] [[2], [1, 3]] [(2, 3)]
            Polynomial : -19/24
            sage: (X.xi^X.dim()).evaluate() == (X.xi^X.dim()).to_prodtautclass().pushforward().evaluate()
            True
        """
        G = self._X.smooth_LG
        stgraph = G.LG.stgraph
        total = admcycles.admcycles.prodtautclass(stgraph, terms=[])
        for c, AG in self.psi_list:
            ptc = AG.to_prodtautclass()
            # sort vertices by connected component:
            vertex_map = {}
            # note that every vertex of G has at least one leg (that is a
            # marked point of _X):
            for v, _ in enumerate(G.LG.genera):
                mp_on_stratum = G.dmp[G.LG.legs[v][0]]
                # find this marked point on AG:
                l_AG = AG._G.dmp_inv[mp_on_stratum]
                # get the corresponding vertex
                LG = AG._G.LG
                v_AG = LG.vertex(l_AG)
                # we use the underlying graph:
                UG_v = LG.UG_vertex(v_AG)
                for w, g, kind in LG.UG_without_infinity().connected_component_containing_vertex(
                        UG_v):
                    if kind != 'LG':
                        continue
                    vertex_map[w] = v
            # map legs of AG._G to smooth_LG
            # CAREFUL: This goes in the OTHER direction!
            leg_map = {G.dmp_inv[mp]: ldeg for ldeg, mp in AG._G.dmp.items()}
            pf = ptc.partial_pushforward(stgraph, vertex_map, leg_map)
            total += c * pf
        return total
